# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: proto/prediction.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'proto/prediction.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from seldon_core.proto.tensorflow.core.framework import tensor_pb2 as tensorflow_dot_core_dot_framework_dot_tensor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16proto/prediction.proto\x12\rseldon.protos\x1a\x19google/protobuf/any.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a&tensorflow/core/framework/tensor.proto\"\x91\x02\n\rSeldonMessage\x12%\n\x06status\x18\x01 \x01(\x0b\x32\x15.seldon.protos.Status\x12!\n\x04meta\x18\x02 \x01(\x0b\x32\x13.seldon.protos.Meta\x12*\n\x04\x64\x61ta\x18\x03 \x01(\x0b\x32\x1a.seldon.protos.DefaultDataH\x00\x12\x11\n\x07\x62inData\x18\x04 \x01(\x0cH\x00\x12\x11\n\x07strData\x18\x05 \x01(\tH\x00\x12*\n\x08jsonData\x18\x06 \x01(\x0b\x32\x16.google.protobuf.ValueH\x00\x12*\n\ncustomData\x18\x07 \x01(\x0b\x32\x14.google.protobuf.AnyH\x00\x42\x0c\n\ndata_oneof\"\xaf\x01\n\x0b\x44\x65\x66\x61ultData\x12\r\n\x05names\x18\x01 \x03(\t\x12\'\n\x06tensor\x18\x02 \x01(\x0b\x32\x15.seldon.protos.TensorH\x00\x12-\n\x07ndarray\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.ListValueH\x00\x12+\n\x08tftensor\x18\x04 \x01(\x0b\x32\x17.tensorflow.TensorProtoH\x00\x42\x0c\n\ndata_oneof\"/\n\x06Tensor\x12\x11\n\x05shape\x18\x01 \x03(\x05\x42\x02\x10\x01\x12\x12\n\x06values\x18\x02 \x03(\x01\x42\x02\x10\x01\"\x80\x03\n\x04Meta\x12\x0c\n\x04puid\x18\x01 \x01(\t\x12+\n\x04tags\x18\x02 \x03(\x0b\x32\x1d.seldon.protos.Meta.TagsEntry\x12\x31\n\x07routing\x18\x03 \x03(\x0b\x32 .seldon.protos.Meta.RoutingEntry\x12\x39\n\x0brequestPath\x18\x04 \x03(\x0b\x32$.seldon.protos.Meta.RequestPathEntry\x12&\n\x07metrics\x18\x05 \x03(\x0b\x32\x15.seldon.protos.Metric\x1a\x43\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12%\n\x05value\x18\x02 \x01(\x0b\x32\x16.google.protobuf.Value:\x02\x38\x01\x1a.\n\x0cRoutingEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a\x32\n\x10RequestPathEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xe1\x01\n\x06Metric\x12\x0b\n\x03key\x18\x01 \x01(\t\x12.\n\x04type\x18\x02 \x01(\x0e\x32 .seldon.protos.Metric.MetricType\x12\r\n\x05value\x18\x03 \x01(\x02\x12-\n\x04tags\x18\x04 \x03(\x0b\x32\x1f.seldon.protos.Metric.TagsEntry\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"/\n\nMetricType\x12\x0b\n\x07\x43OUNTER\x10\x00\x12\t\n\x05GAUGE\x10\x01\x12\t\n\x05TIMER\x10\x02\"I\n\x11SeldonMessageList\x12\x34\n\x0eseldonMessages\x18\x01 \x03(\x0b\x32\x1c.seldon.protos.SeldonMessage\"\x8e\x01\n\x06Status\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0c\n\x04info\x18\x02 \x01(\t\x12\x0e\n\x06reason\x18\x03 \x01(\t\x12\x30\n\x06status\x18\x04 \x01(\x0e\x32 .seldon.protos.Status.StatusFlag\"&\n\nStatusFlag\x12\x0b\n\x07SUCCESS\x10\x00\x12\x0b\n\x07\x46\x41ILURE\x10\x01\"\xa6\x01\n\x08\x46\x65\x65\x64\x62\x61\x63k\x12-\n\x07request\x18\x01 \x01(\x0b\x32\x1c.seldon.protos.SeldonMessage\x12.\n\x08response\x18\x02 \x01(\x0b\x32\x1c.seldon.protos.SeldonMessage\x12\x0e\n\x06reward\x18\x03 \x01(\x02\x12+\n\x05truth\x18\x04 \x01(\x0b\x32\x1c.seldon.protos.SeldonMessage\"p\n\x0fRequestResponse\x12-\n\x07request\x18\x01 \x01(\x0b\x32\x1c.seldon.protos.SeldonMessage\x12.\n\x08response\x18\x02 \x01(\x0b\x32\x1c.seldon.protos.SeldonMessage\"*\n\x1aSeldonModelMetadataRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x83\x01\n\x15SeldonMessageMetadata\x12\x13\n\x0bmessagetype\x18\x01 \x01(\t\x12&\n\x06schema\x18\x02 \x01(\x0b\x32\x16.google.protobuf.Value\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x10\n\x08\x64\x61tatype\x18\x04 \x01(\t\x12\r\n\x05shape\x18\x05 \x03(\x03\"\xa3\x02\n\x13SeldonModelMetadata\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08versions\x18\x02 \x03(\t\x12\x10\n\x08platform\x18\x03 \x01(\t\x12\x34\n\x06inputs\x18\x04 \x03(\x0b\x32$.seldon.protos.SeldonMessageMetadata\x12\x35\n\x07outputs\x18\x05 \x03(\x0b\x32$.seldon.protos.SeldonMessageMetadata\x12>\n\x06\x63ustom\x18\x06 \x03(\x0b\x32..seldon.protos.SeldonModelMetadata.CustomEntry\x1a-\n\x0b\x43ustomEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xa3\x02\n\x13SeldonGraphMetadata\x12\x0c\n\x04name\x18\x01 \x01(\t\x12>\n\x06models\x18\x02 \x03(\x0b\x32..seldon.protos.SeldonGraphMetadata.ModelsEntry\x12\x34\n\x06inputs\x18\x03 \x03(\x0b\x32$.seldon.protos.SeldonMessageMetadata\x12\x35\n\x07outputs\x18\x04 \x03(\x0b\x32$.seldon.protos.SeldonMessageMetadata\x1aQ\n\x0bModelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x31\n\x05value\x18\x02 \x01(\x0b\x32\".seldon.protos.SeldonModelMetadata:\x02\x38\x01\x32\x89\x03\n\x07Generic\x12N\n\x0eTransformInput\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12O\n\x0fTransformOutput\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12\x45\n\x05Route\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12M\n\tAggregate\x12 .seldon.protos.SeldonMessageList\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12G\n\x0cSendFeedback\x12\x17.seldon.protos.Feedback\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x32\xe3\x01\n\x05Model\x12G\n\x07Predict\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12G\n\x0cSendFeedback\x12\x17.seldon.protos.Feedback\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12H\n\x08Metadata\x12\x16.google.protobuf.Empty\x1a\".seldon.protos.SeldonModelMetadata\"\x00\x32\x98\x01\n\x06Router\x12\x45\n\x05Route\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12G\n\x0cSendFeedback\x12\x17.seldon.protos.Feedback\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x32]\n\x0bTransformer\x12N\n\x0eTransformInput\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x32\x64\n\x11OutputTransformer\x12O\n\x0fTransformOutput\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x32Y\n\x08\x43ombiner\x12M\n\tAggregate\x12 .seldon.protos.SeldonMessageList\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x32\xcb\x02\n\x06Seldon\x12G\n\x07Predict\x12\x1c.seldon.protos.SeldonMessage\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12G\n\x0cSendFeedback\x12\x17.seldon.protos.Feedback\x1a\x1c.seldon.protos.SeldonMessage\"\x00\x12`\n\rModelMetadata\x12).seldon.protos.SeldonModelMetadataRequest\x1a\".seldon.protos.SeldonModelMetadata\"\x00\x12M\n\rGraphMetadata\x12\x16.google.protobuf.Empty\x1a\".seldon.protos.SeldonGraphMetadata\"\x00\x42h\n\x10io.seldon.protosB\x10PredictionProtosZBgithub.com/seldonio/seldon-core/incubating/wrappers/s2i/go/pkg/apib\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'proto.prediction_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020io.seldon.protosB\020PredictionProtosZBgithub.com/seldonio/seldon-core/incubating/wrappers/s2i/go/pkg/api'
  _globals['_TENSOR'].fields_by_name['shape']._loaded_options = None
  _globals['_TENSOR'].fields_by_name['shape']._serialized_options = b'\020\001'
  _globals['_TENSOR'].fields_by_name['values']._loaded_options = None
  _globals['_TENSOR'].fields_by_name['values']._serialized_options = b'\020\001'
  _globals['_META_TAGSENTRY']._loaded_options = None
  _globals['_META_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_META_ROUTINGENTRY']._loaded_options = None
  _globals['_META_ROUTINGENTRY']._serialized_options = b'8\001'
  _globals['_META_REQUESTPATHENTRY']._loaded_options = None
  _globals['_META_REQUESTPATHENTRY']._serialized_options = b'8\001'
  _globals['_METRIC_TAGSENTRY']._loaded_options = None
  _globals['_METRIC_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_SELDONMODELMETADATA_CUSTOMENTRY']._loaded_options = None
  _globals['_SELDONMODELMETADATA_CUSTOMENTRY']._serialized_options = b'8\001'
  _globals['_SELDONGRAPHMETADATA_MODELSENTRY']._loaded_options = None
  _globals['_SELDONGRAPHMETADATA_MODELSENTRY']._serialized_options = b'8\001'
  _globals['_SELDONMESSAGE']._serialized_start=168
  _globals['_SELDONMESSAGE']._serialized_end=441
  _globals['_DEFAULTDATA']._serialized_start=444
  _globals['_DEFAULTDATA']._serialized_end=619
  _globals['_TENSOR']._serialized_start=621
  _globals['_TENSOR']._serialized_end=668
  _globals['_META']._serialized_start=671
  _globals['_META']._serialized_end=1055
  _globals['_META_TAGSENTRY']._serialized_start=888
  _globals['_META_TAGSENTRY']._serialized_end=955
  _globals['_META_ROUTINGENTRY']._serialized_start=957
  _globals['_META_ROUTINGENTRY']._serialized_end=1003
  _globals['_META_REQUESTPATHENTRY']._serialized_start=1005
  _globals['_META_REQUESTPATHENTRY']._serialized_end=1055
  _globals['_METRIC']._serialized_start=1058
  _globals['_METRIC']._serialized_end=1283
  _globals['_METRIC_TAGSENTRY']._serialized_start=1191
  _globals['_METRIC_TAGSENTRY']._serialized_end=1234
  _globals['_METRIC_METRICTYPE']._serialized_start=1236
  _globals['_METRIC_METRICTYPE']._serialized_end=1283
  _globals['_SELDONMESSAGELIST']._serialized_start=1285
  _globals['_SELDONMESSAGELIST']._serialized_end=1358
  _globals['_STATUS']._serialized_start=1361
  _globals['_STATUS']._serialized_end=1503
  _globals['_STATUS_STATUSFLAG']._serialized_start=1465
  _globals['_STATUS_STATUSFLAG']._serialized_end=1503
  _globals['_FEEDBACK']._serialized_start=1506
  _globals['_FEEDBACK']._serialized_end=1672
  _globals['_REQUESTRESPONSE']._serialized_start=1674
  _globals['_REQUESTRESPONSE']._serialized_end=1786
  _globals['_SELDONMODELMETADATAREQUEST']._serialized_start=1788
  _globals['_SELDONMODELMETADATAREQUEST']._serialized_end=1830
  _globals['_SELDONMESSAGEMETADATA']._serialized_start=1833
  _globals['_SELDONMESSAGEMETADATA']._serialized_end=1964
  _globals['_SELDONMODELMETADATA']._serialized_start=1967
  _globals['_SELDONMODELMETADATA']._serialized_end=2258
  _globals['_SELDONMODELMETADATA_CUSTOMENTRY']._serialized_start=2213
  _globals['_SELDONMODELMETADATA_CUSTOMENTRY']._serialized_end=2258
  _globals['_SELDONGRAPHMETADATA']._serialized_start=2261
  _globals['_SELDONGRAPHMETADATA']._serialized_end=2552
  _globals['_SELDONGRAPHMETADATA_MODELSENTRY']._serialized_start=2471
  _globals['_SELDONGRAPHMETADATA_MODELSENTRY']._serialized_end=2552
  _globals['_GENERIC']._serialized_start=2555
  _globals['_GENERIC']._serialized_end=2948
  _globals['_MODEL']._serialized_start=2951
  _globals['_MODEL']._serialized_end=3178
  _globals['_ROUTER']._serialized_start=3181
  _globals['_ROUTER']._serialized_end=3333
  _globals['_TRANSFORMER']._serialized_start=3335
  _globals['_TRANSFORMER']._serialized_end=3428
  _globals['_OUTPUTTRANSFORMER']._serialized_start=3430
  _globals['_OUTPUTTRANSFORMER']._serialized_end=3530
  _globals['_COMBINER']._serialized_start=3532
  _globals['_COMBINER']._serialized_end=3621
  _globals['_SELDON']._serialized_start=3624
  _globals['_SELDON']._serialized_end=3955
# @@protoc_insertion_point(module_scope)
