# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from seldon_core.proto import prediction_pb2 as proto_dot_prediction__pb2

GRPC_GENERATED_VERSION = '1.70.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in proto/prediction_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class GenericStub(object):
    """[END Messages]

    [START Services]

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.TransformInput = channel.unary_unary(
                '/seldon.protos.Generic/TransformInput',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.TransformOutput = channel.unary_unary(
                '/seldon.protos.Generic/TransformOutput',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.Route = channel.unary_unary(
                '/seldon.protos.Generic/Route',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.Aggregate = channel.unary_unary(
                '/seldon.protos.Generic/Aggregate',
                request_serializer=proto_dot_prediction__pb2.SeldonMessageList.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.SendFeedback = channel.unary_unary(
                '/seldon.protos.Generic/SendFeedback',
                request_serializer=proto_dot_prediction__pb2.Feedback.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)


class GenericServicer(object):
    """[END Messages]

    [START Services]

    """

    def TransformInput(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TransformOutput(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Route(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Aggregate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendFeedback(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GenericServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'TransformInput': grpc.unary_unary_rpc_method_handler(
                    servicer.TransformInput,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'TransformOutput': grpc.unary_unary_rpc_method_handler(
                    servicer.TransformOutput,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'Route': grpc.unary_unary_rpc_method_handler(
                    servicer.Route,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'Aggregate': grpc.unary_unary_rpc_method_handler(
                    servicer.Aggregate,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessageList.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'SendFeedback': grpc.unary_unary_rpc_method_handler(
                    servicer.SendFeedback,
                    request_deserializer=proto_dot_prediction__pb2.Feedback.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'seldon.protos.Generic', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('seldon.protos.Generic', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Generic(object):
    """[END Messages]

    [START Services]

    """

    @staticmethod
    def TransformInput(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Generic/TransformInput',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def TransformOutput(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Generic/TransformOutput',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Route(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Generic/Route',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Aggregate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Generic/Aggregate',
            proto_dot_prediction__pb2.SeldonMessageList.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendFeedback(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Generic/SendFeedback',
            proto_dot_prediction__pb2.Feedback.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class ModelStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Predict = channel.unary_unary(
                '/seldon.protos.Model/Predict',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.SendFeedback = channel.unary_unary(
                '/seldon.protos.Model/SendFeedback',
                request_serializer=proto_dot_prediction__pb2.Feedback.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.Metadata = channel.unary_unary(
                '/seldon.protos.Model/Metadata',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonModelMetadata.FromString,
                _registered_method=True)


class ModelServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Predict(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendFeedback(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Metadata(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ModelServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Predict': grpc.unary_unary_rpc_method_handler(
                    servicer.Predict,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'SendFeedback': grpc.unary_unary_rpc_method_handler(
                    servicer.SendFeedback,
                    request_deserializer=proto_dot_prediction__pb2.Feedback.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'Metadata': grpc.unary_unary_rpc_method_handler(
                    servicer.Metadata,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonModelMetadata.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'seldon.protos.Model', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('seldon.protos.Model', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Model(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Predict(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Model/Predict',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendFeedback(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Model/SendFeedback',
            proto_dot_prediction__pb2.Feedback.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Metadata(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Model/Metadata',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            proto_dot_prediction__pb2.SeldonModelMetadata.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class RouterStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Route = channel.unary_unary(
                '/seldon.protos.Router/Route',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.SendFeedback = channel.unary_unary(
                '/seldon.protos.Router/SendFeedback',
                request_serializer=proto_dot_prediction__pb2.Feedback.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)


class RouterServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Route(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendFeedback(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_RouterServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Route': grpc.unary_unary_rpc_method_handler(
                    servicer.Route,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'SendFeedback': grpc.unary_unary_rpc_method_handler(
                    servicer.SendFeedback,
                    request_deserializer=proto_dot_prediction__pb2.Feedback.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'seldon.protos.Router', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('seldon.protos.Router', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Router(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Route(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Router/Route',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendFeedback(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Router/SendFeedback',
            proto_dot_prediction__pb2.Feedback.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class TransformerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.TransformInput = channel.unary_unary(
                '/seldon.protos.Transformer/TransformInput',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)


class TransformerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def TransformInput(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TransformerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'TransformInput': grpc.unary_unary_rpc_method_handler(
                    servicer.TransformInput,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'seldon.protos.Transformer', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('seldon.protos.Transformer', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Transformer(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def TransformInput(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Transformer/TransformInput',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class OutputTransformerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.TransformOutput = channel.unary_unary(
                '/seldon.protos.OutputTransformer/TransformOutput',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)


class OutputTransformerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def TransformOutput(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_OutputTransformerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'TransformOutput': grpc.unary_unary_rpc_method_handler(
                    servicer.TransformOutput,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'seldon.protos.OutputTransformer', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('seldon.protos.OutputTransformer', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class OutputTransformer(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def TransformOutput(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.OutputTransformer/TransformOutput',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class CombinerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Aggregate = channel.unary_unary(
                '/seldon.protos.Combiner/Aggregate',
                request_serializer=proto_dot_prediction__pb2.SeldonMessageList.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)


class CombinerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Aggregate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_CombinerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Aggregate': grpc.unary_unary_rpc_method_handler(
                    servicer.Aggregate,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessageList.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'seldon.protos.Combiner', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('seldon.protos.Combiner', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Combiner(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Aggregate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Combiner/Aggregate',
            proto_dot_prediction__pb2.SeldonMessageList.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class SeldonStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Predict = channel.unary_unary(
                '/seldon.protos.Seldon/Predict',
                request_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.SendFeedback = channel.unary_unary(
                '/seldon.protos.Seldon/SendFeedback',
                request_serializer=proto_dot_prediction__pb2.Feedback.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                _registered_method=True)
        self.ModelMetadata = channel.unary_unary(
                '/seldon.protos.Seldon/ModelMetadata',
                request_serializer=proto_dot_prediction__pb2.SeldonModelMetadataRequest.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonModelMetadata.FromString,
                _registered_method=True)
        self.GraphMetadata = channel.unary_unary(
                '/seldon.protos.Seldon/GraphMetadata',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=proto_dot_prediction__pb2.SeldonGraphMetadata.FromString,
                _registered_method=True)


class SeldonServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Predict(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendFeedback(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ModelMetadata(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GraphMetadata(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SeldonServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Predict': grpc.unary_unary_rpc_method_handler(
                    servicer.Predict,
                    request_deserializer=proto_dot_prediction__pb2.SeldonMessage.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'SendFeedback': grpc.unary_unary_rpc_method_handler(
                    servicer.SendFeedback,
                    request_deserializer=proto_dot_prediction__pb2.Feedback.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            ),
            'ModelMetadata': grpc.unary_unary_rpc_method_handler(
                    servicer.ModelMetadata,
                    request_deserializer=proto_dot_prediction__pb2.SeldonModelMetadataRequest.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonModelMetadata.SerializeToString,
            ),
            'GraphMetadata': grpc.unary_unary_rpc_method_handler(
                    servicer.GraphMetadata,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=proto_dot_prediction__pb2.SeldonGraphMetadata.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'seldon.protos.Seldon', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('seldon.protos.Seldon', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Seldon(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Predict(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Seldon/Predict',
            proto_dot_prediction__pb2.SeldonMessage.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendFeedback(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Seldon/SendFeedback',
            proto_dot_prediction__pb2.Feedback.SerializeToString,
            proto_dot_prediction__pb2.SeldonMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ModelMetadata(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Seldon/ModelMetadata',
            proto_dot_prediction__pb2.SeldonModelMetadataRequest.SerializeToString,
            proto_dot_prediction__pb2.SeldonModelMetadata.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GraphMetadata(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/seldon.protos.Seldon/GraphMetadata',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            proto_dot_prediction__pb2.SeldonGraphMetadata.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)
